@@ -30,7 +30,6 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <errno.h>
-#include <limits.h>
 #if defined (__CYGWIN__) || defined (__MSYS__)
 #include <sys/mman.h>
 #endif
@@ -44,6 +43,8 @@
 
 #include <windows.h>
 
+#include "pechecksum.h"
+
 #if defined(__MSYS__)
 /* MSYS has no strtoull */
 unsigned long long strtoull(const char *, char **, int);
@@ -66,6 +67,8 @@ static WORD coff_characteristics_show;
 static WORD pe_characteristics_set;
 static WORD pe_characteristics_clr;
 static WORD pe_characteristics_show;
+static short checksum_show;
+static short checksum_update; /* 1 = update after changes, 2 = fix always */
 
 typedef struct
 {
@@ -132,6 +135,7 @@ enum {
   SIZEOF_HEAP_RESERVE,
   SIZEOF_HEAP_COMMIT,
   SIZEOF_CYGWIN_HEAP,
+  SIZEOF_TIMESTAMP,
   NUM_SIZEOF_VALUES		/* Keep at the end */
 };
 
@@ -153,7 +157,7 @@ typedef struct {
   ULONG       offset32;
 } sizeof_values_t;
 
-sizeof_values_t sizeof_vals[5] = {
+static sizeof_values_t sizeof_vals[NUM_SIZEOF_VALUES] = {
   { 0, "stack reserve size"      , "bytes", 0, FALSE,
     offsetof (IMAGE_NT_HEADERS64, OptionalHeader.SizeOfStackReserve),
     offsetof (IMAGE_NT_HEADERS32, OptionalHeader.SizeOfStackReserve),
@@ -173,6 +177,10 @@ sizeof_values_t sizeof_vals[5] = {
   { 0, "initial Cygwin heap size", "MB", 0, TRUE,
     offsetof (IMAGE_NT_HEADERS64, OptionalHeader.LoaderFlags),
     offsetof (IMAGE_NT_HEADERS32, OptionalHeader.LoaderFlags),
+  },
+  { 0, "file header timestamp", "seconds", 0, TRUE,
+    offsetof (IMAGE_NT_HEADERS64, FileHeader.TimeDateStamp),
+    offsetof (IMAGE_NT_HEADERS32, FileHeader.TimeDateStamp),
   }
 };
 
@@ -193,11 +201,13 @@ static struct option long_options[] = {
   {"wstrim",       optional_argument, NULL, 'w'},
   {"bigaddr",      optional_argument, NULL, 'l'},
   {"sepdbg",       optional_argument, NULL, 'S'},
+  {"checksum",     optional_argument, NULL, 'k'},
   {"stack-reserve",optional_argument, NULL, 'x'},
   {"stack-commit", optional_argument, NULL, 'X'},
   {"heap-reserve", optional_argument, NULL, 'y'},
   {"heap-commit",  optional_argument, NULL, 'Y'},
   {"cygwin-heap",  optional_argument, NULL, 'z'},
+  {"timestamp",    optional_argument, NULL, 'p'},
   {"filelist",     no_argument, NULL, 'T'},
   {"verbose",      no_argument, NULL, 'v'},
   {"help",         no_argument, NULL, 'h'},
@@ -205,7 +215,7 @@ static struct option long_options[] = {
   {NULL, no_argument, NULL, 0}
 };
 static const char *short_options
-	= "d::e::c::f::n::i::s::b::W::t::w::l::S::x::X::y::Y::z::T:vhV";
+	= "d::e::c::f::n::i::s::b::W::t::w::l::S::k::x::X::y::Y::z::p::T:vhV";
 
 static void short_usage (FILE *f);
 static void help (FILE *f);
@@ -228,6 +238,7 @@ static void display_flags (const char *field_name, const symbolic_flags_t *syms,
 			   WORD new_flag_value);
 static char *symbolic_flags (const symbolic_flags_t *syms, long show, long value);
 static void append_and_decorate (char **str, int is_set, const char *name, int len);
+static void do_checksum (const char *pathname, int indent, int changed, ULONG hdr_checksum);
 static void *xmalloc (size_t num);
 #define XMALLOC(type, num)      ((type *) xmalloc ((num) * sizeof(type)))
 static void handle_coff_flag_option (const char *option_name,
@@ -323,6 +334,14 @@ do_mark (const char *pathname)
   WORD new_coff_characteristics;
   WORD old_pe_characteristics;
   WORD new_pe_characteristics;
+  int indent;
+  union
+  {
+    IMAGE_NT_HEADERS32 ntheader32;
+    IMAGE_NT_HEADERS64 ntheader64;
+  } old_ntheader;
+  int changed;
+  ULONG hdr_checksum;
 
   /* Skip if file does not exist */
   if (access (pathname, F_OK) == -1)
@@ -351,6 +370,12 @@ do_mark (const char *pathname)
       return 0;
     }
 
+  /* Save old contents for modification check below. */
+  if (pep->is_64bit)
+    old_ntheader.ntheader64 = *pep->ntheader64;
+  else
+    old_ntheader.ntheader32 = *pep->ntheader32;
+
   get_characteristics (pep,
 		       &old_coff_characteristics,
 		       &old_pe_characteristics);
@@ -428,13 +453,13 @@ do_mark (const char *pathname)
     }
 
   /* Display characteristics. */
+  indent = 0;
   if (verbose
       || !mark_any
       || coff_characteristics_show
       || pe_characteristics_show
       || handle_any_sizeof != DONT_HANDLE)
     {
-      BOOL printed_characteristic = FALSE;
       int i;
 
       printf ("%s: ", pathname);
@@ -453,27 +478,54 @@ do_mark (const char *pathname)
 			 old_pe_characteristics,
 			 new_pe_characteristics);
 	  puts ("");
-	  printed_characteristic = TRUE;
+	  indent = (int) strlen (pathname) + 2;
 	}
 
       for (i = 0; i < NUM_SIZEOF_VALUES; ++i)
 	{
 	  if (sizeof_vals[i].handle != DONT_HANDLE)
 	    {
-	      printf ("%*s%-24s: %" PRIu64 " (0x%" PRIx64 ") %s\n",
-		      printed_characteristic ? (int) strlen (pathname) + 2
-					     : 0, "",
+              char extra[32] = "";
+	      if (verbose && i == SIZEOF_TIMESTAMP)
+		{
+		  time_t t = (time_t)sizeof_vals[i].value;
+		  const struct tm * tp = gmtime (&t);
+		  if (tp)
+		    strftime (extra, sizeof(extra), " [%Y-%m-%d %H:%M:%S UTC]", tp);
+		}
+
+	      printf ("%*s%-24s: %" PRIu64 " (0x%" PRIx64 ") %s%s\n",
+		      indent, "",
 		      sizeof_vals[i].name,
 		      (uint64_t) sizeof_vals[i].value,
 		      (uint64_t) sizeof_vals[i].value,
-		      sizeof_vals[i].unit);
-	      printed_characteristic = TRUE;
+		      sizeof_vals[i].unit, extra);
+	      if (!indent)
+		indent = (int) strlen (pathname) + 2;
 	    }
 	}
+
+    }
+
+  /* Any actual changes? */
+  if (pep->is_64bit)
+    {
+      changed = memcmp(&old_ntheader.ntheader64, pep->ntheader64,
+		       sizeof(old_ntheader.ntheader64));
+      hdr_checksum = pep->ntheader64->OptionalHeader.CheckSum;
+    }
+  else
+    {
+      changed = memcmp(&old_ntheader.ntheader32, pep->ntheader64,
+		       sizeof(old_ntheader.ntheader32));
+      hdr_checksum = pep->ntheader32->OptionalHeader.CheckSum;
     }
 
   pe_close (pep);
 
+  /* Checksum calculation requires re-open as a regular file. */
+  if (checksum_show || checksum_update)
+    do_checksum (pathname, indent, changed, hdr_checksum);
   return 0;
 }
 
@@ -569,6 +621,51 @@ append_and_decorate (char **str, int is_set, const char *name, int len)
     } 
 }
 
+static void
+do_checksum (const char *pathname, int indent, int changed, ULONG hdr_checksum)
+{
+  const char name[] = "PE file checksum";
+  if (checksum_show || (checksum_update && changed) || checksum_update > 1)
+    {
+      int fd;
+      unsigned new_checksum, old_checksum;
+      if ((fd = open (pathname, (checksum_update ? O_RDWR : O_RDONLY) | O_BINARY)) == -1)
+	{
+	  fprintf (stderr, "%s: Unable to reopen, errno=%d\n", pathname, errno);
+	  return;
+	}
+      new_checksum = pe32_checksum (fd, checksum_update, &old_checksum);
+      close(fd);
+      if (!new_checksum)
+	{
+	  fprintf (stderr, "%s: Checksum calculation failed, errno=%d\n",
+		   pathname, errno);
+	  return;
+	}
+      if (old_checksum != hdr_checksum)
+	{
+	  fprintf (stderr, "%s: Internal error: Re-read checksum does not match: "
+		  "0x%08x != 0x%08x\n",  pathname, old_checksum, hdr_checksum);
+	  return;
+	}
+
+      if (new_checksum == old_checksum)
+	  printf ("%*s%-24s: 0x%08x (%sverified)\n", indent, "", name,
+		  old_checksum, (checksum_update ? "unchanged, " : ""));
+      else if (checksum_update)
+	  printf ("%*s%-24s: 0x%08x (updated from 0x%08x)\n", indent, "", name,
+		  new_checksum, old_checksum);
+      else
+	  printf ("%*s%-24s: 0x%08x (*needs update to 0x%08x*)\n", indent, "", name,
+		  old_checksum, new_checksum);
+    }
+  else /* (!checksum_show && checksum_update == 1 && !changed) */
+    {
+      printf ("%*s%-24s: 0x%08x (unchanged, not verified)\n", indent, "", name,
+	      hdr_checksum);
+    }
+}
+
 static void *
 xmalloc (size_t num)
 {
@@ -598,7 +695,7 @@ handle_num_option (const char *option_name,
 	   || sizeof_vals[option_index].value > 0x0000ffffffffffffULL
 	   /* Just a ULONG value */
 	   || (sizeof_vals[option_index].is_ulong
-	       && sizeof_vals[option_index].value > ULONG_MAX))
+	       && sizeof_vals[option_index].value > MAXULONG32))
     {
       fprintf (stderr, "Invalid argument for %s: %s\n", 
 	       option_name, option_arg);
@@ -664,6 +761,31 @@ handle_coff_flag_option (const char *option_name,
     }
 }
 
+static void
+handle_checksum_option (const char *option_name,
+                        const char *option_arg)
+{
+  int bool_value;
+  if (!option_arg)
+    {
+      checksum_show = 1;
+      if (handle_any_sizeof == DONT_HANDLE)
+	handle_any_sizeof = DO_READ;
+    }
+  else
+    {
+      if (string_to_bool (option_arg, &bool_value) != 0)
+        {
+          fprintf (stderr, "Invalid argument for %s: %s\n",
+                   option_name, option_arg);
+          short_usage (stderr);
+          exit (1);
+        }
+      checksum_update = (bool_value ? 2 : 1);
+      handle_any_sizeof = DO_WRITE;
+    }
+}
+
 void
 parse_args (int argc, char *argv[])
 {
@@ -762,6 +884,10 @@ parse_args (int argc, char *argv[])
 	                           optarg,
 	                           IMAGE_FILE_DEBUG_STRIPPED);
 	  break;
+	case 'k':
+	  handle_checksum_option (long_options[option_index].name,
+				  optarg);
+	  break;
 	case 'x':
 	  handle_num_option (long_options[option_index].name,
 			     optarg,
@@ -787,6 +913,11 @@ parse_args (int argc, char *argv[])
 			     optarg,
 			     SIZEOF_CYGWIN_HEAP);
 	  break;
+	case 'p':
+	  handle_num_option (long_options[option_index].name,
+			     optarg,
+			     SIZEOF_TIMESTAMP);
+	  break;
 	case 'T':
 	  file_list = optarg;
 	  break;
@@ -960,7 +1091,7 @@ get_and_set_size (const pe_file *pep, sizeof_values_t *val)
     }
   else if (val->handle == DO_WRITE)
     {
-      if ((!pep->is_64bit || val->is_ulong) && val->value >= ULONG_MAX)
+      if ((!pep->is_64bit || val->is_ulong) && val->value > MAXULONG32)
 	{
 	  fprintf (stderr, "%s: Skip writing %s, value too big\n",
 		   pep->pathname, val->name);
@@ -1102,6 +1233,9 @@ help (FILE *f)
 "                              than 2 GB.\n"
 "  -S, --sepdbg       [BOOL]   Debugging information was removed and stored\n"
 "                              separately in another file.\n"
+"  -k, --checksum     [BOOL]   Displays (no argument), updates after changes\n"
+"                              only (false) or always fixes (true) the file\n"
+"                              checksum in the PE header.\n"
 "  -x, --stack-reserve [NUM]   Reserved stack size of the process in bytes.\n"
 "  -X, --stack-commit  [NUM]   Initial commited portion of the process stack\n"
 "                              in bytes.\n"
@@ -1119,6 +1253,7 @@ help (FILE *f)
 "                              Useful values are between 4 and 2048.  If 0,\n"
 "                              Cygwin uses the default heap size of 384 Megs.\n"
 "                              Has no meaning for non-Cygwin applications.\n"
+"  -p, --timestamp     [NUM]   Timestamp in file header (seconds since epoch).\n"
 "  -T, --filelist FILE         Indicate that FILE contains a list\n"
 "                              of PE files to process\n"
 "  -v, --verbose               Display diagnostic information\n"
@@ -1136,6 +1271,8 @@ help (FILE *f)
 "                                --cygwin-heap, --cygwin-heap=512, etc\n"
 "For flag values, to set a value, and display the results symbolic, repeat the\n"
 "option:  --tsaware=true --tsaware -d0 -d\n"
+"The time of last modification of each file is preserved unless the checksum\n"
+"has been updated.\n"
 "\n", f);
 }
 
